module 0-Logic where 

-- In this practical we play around with Curry-Howard
-- isomorphism in Agda.

variable
  P Q R : Set
  A B C : Set

-- Implication is encoded by function space.

I : P → P
I = λ x → x 
-- or I x = x 

K : P → (Q → P)
K = λ x → λ y → x
-- or K = λ x y → x
-- or K x y = x

S : (P → Q → R) → (P → Q) → P → R
S = λ f → λ g → λ x → f x (g x)
-- or S f g x = f x (g x)

flip : (P → Q → R) → Q → P → R
flip f y x = f x y 

prac7_1d : P → (P → Q) → (P → Q → R) → R
prac7_1d x f g = g x (f x)

-- True is encoded by the type ⊤, a type with exactly
-- one component tt.

data ⊤ : Set where
  tt : ⊤

-- False is represented by ⊥, a type having no terms.

data ⊥ : Set where

-- Negation of P is P → False.

¬ : Set → Set
¬ P = P → ⊥

-- Conjunction P ∧ Q is encoded by a pair P × Q.

data _∧_ (A B : Set) : Set where
  _,_ : A → B → A ∧ B

infixr 4 _,_
infixr 2 _∧_

fst : A ∧ B → A
fst (x , y) = x

snd : A ∧ B → B
snd (x , y) = y

-- Disjunction P ∨ Q is encoded by the sum type.

data _∨_ (A B : Set) : Set where
  inl : A → A ∨ B
  inr : B → A ∨ B

infixr 1 _∨_

-- Exercises from your Logic class!

uncurry : (P → Q → R) → (P ∧ Q) → R
uncurry f (x , y) = f x y

∨-comm : (P ∨ Q) → (Q ∨ P)
∨-comm (inl p) = inr p
∨-comm (inr q) = inl q

  -- ∨ "\ v e e"      "\ o r"
  -- ∧ "\ w e d g e"  "\ a n d"

∧-comm : (P ∧ Q) → (Q ∧ P)
∧-comm (p , q) = (q , p)

→-∨-weakening-r : (A → B) → (A → (B ∨ C))
→-∨-weakening-r f x = inl (f x)

→-∨-weakening-l : ((A ∨ C) → B) → (A → B)
→-∨-weakening-l f x = f (inl x)

→-∧-weakening-r1 : (A → (B ∧ C)) → (A → B)
→-∧-weakening-r1 f x = fst (f x)

→-∧-weakening-r2 : (A → (B ∧ C)) → (A → C)
→-∧-weakening-r2 f x = snd (f x)

→-∧-distr : (A → (B ∧ C)) → ((A → B) ∧ (A → C))
→-∧-distr f = (λ x → fst (f x)) , (λ x → snd (f x))
-- OR →-∧-distr f = →-∧-weakening-r1 f , →-∧-weakening-r2 f

⊥-elim : ⊥ → A
⊥-elim ()

resol : ((A ∨ B) ∧ ¬ B) → A
resol (inl x , f) = x
resol (inr y , f) = ⊥-elim (f y)

prac7_1f : (P ∧ Q) → ((P ∨ Q) → R) → R
prac7_1f (p , q) f = f (inl p)
-- OR prac7_1f (p , q) f = f (inr q)

¬¬ex-middle : ¬ (¬ (A ∨ ¬ A))
¬¬ex-middle f = f (inr (λ a → f (inl a)))

{- However, we cannot prove that:

ex-middle : A ∨ (¬ A)
ex-middle = ?
-}

A→¬¬A : A → ¬ (¬ A)
A→¬¬A a f = f a


{- However, we cannot prove that:

¬¬A→A : ¬ (¬ A) → A
¬¬A→A = ?
-}

demorgan : (¬ A ∨ ¬ B) → ¬ (A ∧ B)
demorgan (inl na) (a , b) = ⊥-elim (na a)
demorgan (inr nb) (a , b) = ⊥-elim (nb b)

{- However, we cannot prove that

demorgan' : ¬ (A ∧ B) → (¬ A ∨ ¬ B)
demorgan' = ?

-}

contra : (A → B) → (¬ B → ¬ A)
contra f nb a = ⊥-elim (nb (f a))

{- However, we cannot prove that

contra' : (¬ B → ¬ A) → (A → B)
contra' = ?
-}


absorb : (A ∧ (¬ A ∨ B)) → (A ∧ B)
absorb (a , inl na) = ⊥-elim (na a)
absorb (a , inr b) = (a , b)

∨-∧-distr : (A ∨ (B ∧ C)) → (A ∨ B) ∧ (A ∨ C)
∨-∧-distr (inl a) = (inl a , inl a)
∨-∧-distr (inr (b , c)) = (inr b , inr c)
